import pandas as pd

from sklearn.ensemble import RandomForestRegressor
from sklearn.metrics import mean_squared_error, r2_score
from sklearn.neighbors import KNeighborsRegressor
from sklearn.neural_network import MLPRegressor
from sklearn.linear_model import Ridge
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.model_selection import cross_val_predict, KFold, RandomizedSearchCV
from joblib import dump, load
from helpers.visualization_helpers import get_name
from scipy.stats import uniform

import csv

# Models
rf = RandomForestRegressor(n_estimators=750,
                           max_depth=15,
                           min_samples_split=75,
                           max_features='log2',
                           random_state=42,
                           bootstrap=True,
                           n_jobs=-1,
                           verbose=0)
knn = KNeighborsRegressor(n_neighbors=2, p=2, n_jobs=-1)
mlp = MLPRegressor(hidden_layer_sizes=(100,),  # 1 layer with 100 units
                   activation='relu',          # ReLU activation function
                   solver='adam',              # Adam optimizer
                   random_state=42,
                   verbose=0)
ridge = Ridge(alpha=1, random_state=42)
gbf = GradientBoostingRegressor(n_estimators=200,
                                min_samples_split=500,
                                min_samples_leaf=50,
                                learning_rate=0.1,
                                max_features='sqrt',
                                subsample=0.8,
                                max_depth=10, 
                                random_state=42,
                                verbose=0)

model_params_rf = {
    'n_estimators': [250, 500, 750, 1000],
    'max_features': ['sqrt', 'log2', 'auto'],
    'max_depth': [3, 6, 9],
    'max_leaf_nodes': [3, 6, 9],
    'min_samples_split': [50, 75, 100],
    'bootstrap': [True, False],
}

model_params_knn = {
    'n_neighbors': [1, 2, 3, 4, 5, 6, 7, 8],
}

model_params_mlp = {
    'hidden_layer_sizes': [75, 100, 125],
    'activation': ['tanh', 'relu', 'logistic'],
    'solver': ['sgd', 'adam', 'lbfgs'],
    'alpha': uniform(0.0001, 0.9),
    'learning_rate': ['constant','adaptive'],
}

model_params_ridge = {
    'alpha': [1, 2, 3],
}

model_params_gb = {
    'n_estimators': [100, 200, 500, 1000],
    'min_samples_split': [250, 500, 1000],
    'min_samples_leaf': [10, 20, 50],
    'learning_rate': [0.05, 0.1, 0.5, 1],
    'subsample' : [0.5, 0.8, 1],
}

params = [model_params_rf, model_params_knn, model_params_mlp, model_params_ridge, model_params_gb]

def get_y(data):
    return data['CostUSD']

def dummy(data):
    # 'CostUSD' is the target variable
    X = data.drop(columns=['CostUSD','Cost', 'tags', 'ResourceId'])

    # Handle missing values if any
    X = X.fillna(method='ffill').fillna(method='bfill')
    
    n_columns_to_process = 9  # Choose the number of columns to process
    top_columns = X.nunique().nlargest(n_columns_to_process).index
    
    top_n_categories_for_columns = 20

    for column in top_columns:
        if X[column].dtype == 'object':
            # Select only the top N most frequent categories and group the rest as 'Other'
            column_counts = X[column].value_counts()
            column_to_keep = column_counts[:top_n_categories_for_columns].index
            X[column] = X[column].apply(lambda x: x if x in column_to_keep else 'Other')

    # Create dummy variables for categorical features
    X = pd.get_dummies(X, drop_first=True)
    
    return X

# Trains and tests models for the multi-model approach
def cross_val_models(df_types, models, types, n_splits=5, tune=False):
    if tune:
            f = open("results/hyperparameters.csv", "a")
            f.write('ResourceType,Model,MSE,R2,Diff%,params\n')
            f.close()
    
    else:      
            f = open("results/cross_validation.csv", "a")
            f.write("ResourceType,Model,MSE,R2,Diff%\n")
            f.close()

    print('-------Beginning cross-validation training and testing-------')
    
    for i in range(len(df_types)):
        X = dummy(df_types[i])
        y = get_y(df_types[i])
        
        if len(df_types[i].index) < 50:
            print(f"Too few datapoints for index {i}.")
        else:
            for j in range(len(models)):
                model = models[j]
                kf = KFold(n_splits=n_splits, shuffle=True, random_state=42)
                
                mse_scores = []
                r2_scores = []
                diff_scores = []
                
                for train_index, test_index in kf.split(X):
                    X_train, X_test = X.iloc[train_index], X.iloc[test_index]
                    y_train, y_test = y.iloc[train_index], y.iloc[test_index]
                    
                    if tune:
                        randm_src = RandomizedSearchCV(estimator=models[j], param_distributions=params[j], cv=2, n_iter=10, n_jobs=-1)
                        randm_src.fit(X_train, y_train)
                        y_pred = randm_src.predict(X_test)

                    else:
                        model.fit(X_train, y_train)
                        y_pred = model.predict(X_test)

                    mse = mean_squared_error(y_test, y_pred)
                    r2 = r2_score(y_test, y_pred)
                    
                    avg_test = get_avg(y_test)
                    avg_pred = get_avg(y_pred)
                    diff = (abs(avg_pred - avg_test) / avg_test) * 100
                    
                    mse_scores.append(mse)
                    r2_scores.append(r2)
                    diff_scores.append(diff)
                
                avg_mse = sum(mse_scores) / n_splits
                avg_r2 = sum(r2_scores) / n_splits
                avg_diff = sum(diff_scores) / n_splits
                
                if tune:
                        f = open("results/hyperparameters.csv", "a")
                        f.write(f'{types[i]},{type(model).__name__},{avg_mse},{avg_r2},{avg_diff},{str(randm_src.best_params_)}\n')
                        f.close()

                        dump(randm_src.best_estimator_, f'models/{get_name(resource_type=types[i], model=type(randm_src.best_estimator_).__name__)}.joblib')
                
                else:
                        f = open("results/cross_validation.csv", "a")
                        f.write(f"{types[i]},{type(model).__name__},{avg_mse},{avg_r2},{avg_diff}\n")
                        f.close()

                        dump(model, f'models/{get_name(model=type(model).__name__, resource_type=types[i])}')
    
    return

# Trains and tests the model for the monolithic approach.
def cross_val_monolith(data, model, n_splits=5, tune=False):
    f = open("results/monolith.csv", "a")
    f.write("ResourceType,Model,MSE,R2,Diff%\n")
    f.close()

    X = dummy(data)
    y = get_y(data)
    
    kf = KFold(n_splits=n_splits, shuffle=True, random_state=42)
    
    mse_scores = []
    r2_scores = []
    diff_scores = []
    
    for train_index, test_index in kf.split(X):
        X_train, X_test = X.iloc[train_index], X.iloc[test_index]
        y_train, y_test = y.iloc[train_index], y.iloc[test_index]
        
        if tune:
            randm_src = RandomizedSearchCV(estimator=model, param_distributions=params[0], cv=2, n_iter=10, n_jobs=-1)
            randm_src.fit(X_train, y_train)
            y_pred = randm_src.predict(X_test)

        else:
            model = model.fit(X_train, y_train)
            y_pred = model.predict(X_test)
        
        mse = mean_squared_error(y_test, y_pred)
        r2 = r2_score(y_test, y_pred)
        
        avg_test = get_avg(y_test)
        avg_pred = get_avg(y_pred)
        diff = (abs(avg_pred - avg_test) / avg_test) * 100
        
        mse_scores.append(mse)
        r2_scores.append(r2)
        diff_scores.append(diff)
    
    avg_mse = sum(mse_scores) / n_splits
    avg_r2 = sum(r2_scores) / n_splits
    avg_diff = sum(diff_scores) / n_splits
    
    if tune:
            f = open("results/cross_validation.csv", "a")
            f.write(f'monolith,{type(model).__name__},{avg_mse},{avg_r2},{avg_diff},{str(randm_src.best_params_)}\n')
            f.close()
    else:
            f = open("results/monolith.csv", "a")
            f.write(f"'monolith,{type(model).__name__},{avg_mse},{avg_r2},{avg_diff}\n")
            f.close()

    dump(model, 'models/monolith.joblib')

    return

def get_models():
    model_list = [rf, knn, mlp, ridge, gbf]
    return model_list
            

def load_best_models(filename):
    file = open(filename)
    csvreader = csv.reader(file)

    header = []
    header = next(csvreader)

    loaded_models = []
    properties = []
    for row in csvreader:
        model = load(f'models/{get_name(resource_type=row[0],model=row[1])}.joblib')
        loaded_models.append(model)
        properties.append([row[0], row[1]])

    return loaded_models, properties

def get_avg(list):
    sum = 0
    for i in list:
        sum+=i
    return sum/len(list)

def get_type_dfs(types, columns):
    dfs = []
    for type in types:
        df = pd.DataFrame(columns=columns)
        dfs.append(df)
    return dfs

def get_date_indexes(df_arr):
    df_arr = update_indexes(df_arr)
    mapping = {}
    for i, df in enumerate(df_arr):
        for old_index in df.index:
            mapping[old_index] = i

    return mapping

def update_indexes(df_arr):
    updated_dataframes_list = []
    offset = 0
    
    for df in df_arr:
        new_indexes = df.index + offset
        df.index = new_indexes
        updated_dataframes_list.append(df)
        
        # Update the offset based on the length of the current DataFrame
        offset += len(df)
        
    return updated_dataframes_list
            

def append_pred(model, data):
    X = dummy(data)
    y = get_y(data)
    #X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.3, random_state=42)
    y_pred = cross_val_predict(model, X, y, cv=5)
    X['pred'] = y_pred
    X['CostUSD'] = y

    return X

def find_df(resource_type, df_types):
    for df in df_types:
        if df['ResourceType'].values[0] == resource_type:
            return df
        
    return -1